#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "blackScholes.h"
#include "control3.h"
#include "util.h"

#define WEIGHT(time,prices,toPath) transProb(&(md->option.pData),prices,md->option.pData.price[time+1][toPath]) /md->distribution[(time)][(toPath)]

/* The exact value of control 1 used for the american estimate at time, path */
double innerControl1Exact(meshData *md, int time, double *prices){
  int max;

  max = dWhichMax(md->option.pData.underliers, prices);
  return(exp(-time*md->option.riskFreeRate)*
         euroCall(prices[max],
                  md->option.strikePrice,
                  md->option.riskFreeRate,
                  sqrt(md->option.pData.covariance[max][max]),
                  1.0,
                  md->option.pData.dividendRate[max]));
}

/* The mesh estimate of control1 at time, path */
double innerControl1Estimate(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r;
  
  max = dWhichMax(md->option.pData.underliers, prices); 
  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += WEIGHT(time,prices,toPath);
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*
        (md->option.pData.price[time+1][toPath][max] - md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices,toPath)*deflatedPrice;
  }
  estimate /= sumWeight;
  return(exp(-time*md->option.riskFreeRate)*estimate);
}

/* The variance of the mesh estimate of control1 at time,path */
double innerControl1Variance(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,sumWeight,deflatedPrice;

  max = dWhichMax(md->option.pData.underliers, prices); 

  estimate = 0;
  sumWeight =0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*
        (md->option.pData.price[time+1][toPath][max]-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices,toPath) *deflatedPrice*deflatedPrice;
    sumWeight += WEIGHT(time,prices,toPath);
  }
  estimate /= sumWeight;
  return(exp(-2*time*md->option.riskFreeRate)*estimate);
}

/* The correlation of the american estimate and control1 */
double innerControl1Correlation(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,sumWeight,deflatedPrice;

  max = dWhichMax(md->option.pData.underliers, prices); 

  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*
        (md->option.pData.price[time+1][toPath][max]-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices,toPath)*exp(-time*md->option.riskFreeRate)*
      deflatedPrice*md->newValue[time+1][toPath];
    sumWeight += WEIGHT(time,prices,toPath);
  }
  estimate /= sumWeight;
  return(estimate);
} 







double innerControl2Exact(meshData *md, int time, double *prices){
  int max;

  max = dWhichMax(md->option.pData.underliers, prices);
  return(exp(md->option.pData.growthRate[max]-
             md->option.pData.dividendRate[max])*prices[max]);
}

double innerControl2Estimate(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate, sumWeight;

  max = dWhichMax(md->option.pData.underliers, prices);
  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath < md->meshPaths; toPath++){
    sumWeight += WEIGHT(time,prices,toPath);
    estimate += WEIGHT(time,prices,toPath)*
      md->option.pData.price[time+1][toPath][max];
  }
  estimate /= sumWeight;
  return(estimate);
}


double innerControl2Variance(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,sumWeight;

  max = dWhichMax(md->option.pData.underliers, prices); 
  estimate = 0;
  sumWeight =0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    estimate += WEIGHT(time,prices,toPath)*prices[max]*prices[max];
    sumWeight += WEIGHT(time,prices,toPath);
  }
  estimate /= sumWeight;
  return(exp(-2*time*md->option.riskFreeRate)*estimate);
}



double innerControl2Correlation(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,sumWeight;

  max = dWhichMax(md->option.pData.underliers, prices); 
  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    estimate += WEIGHT(time,prices,toPath)*
      md->option.pData.price[time+1][toPath][max]*md->newValue[time+1][toPath];
    sumWeight += WEIGHT(time,prices,toPath);
  }
  estimate /= sumWeight;
  return(estimate);
} 



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Control 3 estimators and exact calculations */

/* The following computes the exact value of control 3 using equations from
   Stulz. See control3.c */

double innerControl3Exact(meshData *md, int time, double *prices){
  int max1, max2, i;
  double maxval1, maxval2;

  /* At the given time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, prices);
  maxval1 = prices[max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && prices[i] > maxval2){
      max2 = i;
      maxval2 = prices[i];
    }
  }

  /* Evaluate deflated value of option price as predicted by Stulz. */
  /*
    return(exp(-time*md->option.riskFreeRate)*  \
    stulzMax(maxval1, maxval2, sqrt(md->option.pData.covariance[max1][max1]), \
    sqrt(md->option.pData.covariance[max2][max2]), \
    md->option.pData.covariance[max1][max2], md->option.riskFreeRate, \
    md->option.strikePrice,
    1.0, md->option.pData.dividendRate[max2])); 
  */
  return(exp(-time*md->option.riskFreeRate)* 
         maxOf2Risk(maxval1, maxval2, md->option.strikePrice, 1.0,
                    sqrt(md->option.pData.covariance[max1][max1]), 
                    sqrt(md->option.pData.covariance[max2][max2]), 
                    md->option.riskFreeRate, 
                    md->option.pData.dividendRate[max1], 
                    md->option.pData.dividendRate[max2],
                    md->option.pData.covariance[max1][max2]));
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* The mesh estimate of control3 at time, path */
double innerControl3Estimate(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* At the current time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, prices);
  maxval1 = prices[max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && prices[i] > maxval2){
      max2 = i;
      maxval2 = prices[i];
    }
  }
  
  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += WEIGHT(time,prices,toPath);
    if( md->option.pData.price[time+1][toPath][max1] >= 
        md->option.pData.price[time+1][toPath][max2]){
      maxPrice = md->option.pData.price[time+1][toPath][max1];
    }else{
      maxPrice = md->option.pData.price[time+1][toPath][max2];
    }
    if(maxPrice > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*
        (maxPrice - md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices,toPath)*deflatedPrice;
  }
  estimate /= sumWeight;
  return(exp(-time*md->option.riskFreeRate)*estimate);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* The variance of the mesh estimate of control3 at time,path */
double innerControl3Variance(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* At the current time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, prices);
  maxval1 = prices[max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && prices[i] > maxval2){
      max2 = i;
      maxval2 = prices[i];
    }
  }
  
  estimate = 0;
  sumWeight =0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if( md->option.pData.price[time+1][toPath][max1] >= 
        md->option.pData.price[time+1][toPath][max2]){
      maxPrice = md->option.pData.price[time+1][toPath][max1];
    }else{
      maxPrice = md->option.pData.price[time+1][toPath][max2];
    }
    if(maxPrice > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*
        (maxPrice-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices,toPath) *deflatedPrice*deflatedPrice;
    sumWeight += WEIGHT(time,prices,toPath);
  }
  estimate /= sumWeight;
  return(exp(-2*time*md->option.riskFreeRate)*estimate);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* The correlation of the american estimate and control3 */
double innerControl3Correlation(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* At the current time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, prices);
  maxval1 = prices[max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && prices[i] > maxval2){
      max2 = i;
      maxval2 = prices[i];
    }
  }
  
  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if( md->option.pData.price[time+1][toPath][max1] >=
        md->option.pData.price[time+1][toPath][max2]){
      maxPrice = md->option.pData.price[time+1][toPath][max1];
    }else{
      maxPrice = md->option.pData.price[time+1][toPath][max2];
    }

    if(maxPrice > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*
        (maxPrice - md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += WEIGHT(time,prices,toPath) * exp(-time*md->option.riskFreeRate)
      *deflatedPrice*md->newValue[time+1][toPath];
    sumWeight += WEIGHT(time,prices,toPath);
  }
  estimate /= sumWeight;
  return(estimate);
} 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */





/* Mesh estimate of american price.  Should probably be in mesh.c */
double innerMeshEstimate(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate,sumWeight;
  double w,v;

  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += WEIGHT(time,prices,toPath);
    estimate +=  WEIGHT(time,prices,toPath)*md->newValue[time+1][toPath];
    /*    printf("M:%i\t%i\t%i\t%lf\t%lf\n",time,path,toPath,
          WEIGHT(time,prices,toPath),md->newValue[time+1][toPath]);*/
  }

  estimate /= sumWeight;
  /*  printf("M:%lf\t%lf\n",sumWeight,estimate);*/
  return(estimate);
} 

double innerOldMeshEstimate(meshData *md, int time, double *prices){
  int max, toPath;
  double estimate;
  double w,v;

  estimate = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    estimate +=  WEIGHT(time,prices,toPath)*md->newValue[time+1][toPath];
  }
  return(estimate);
} 


/* The controlled value of the american estimate at time, path */
double innerNewEstimate(meshData *md, int time, double *prices,
                        int whichControl){
  double beta, I1, J1, J1V, J, I1J1, S;
  int toPath, fromPath;

  S = 0;
  for(toPath=0;toPath<md->meshPaths;toPath++){
    S += WEIGHT(time,prices,toPath);
  }
  switch(whichControl){
  case(0):
    return(innerOldMeshEstimate(md,time,prices));
    break;
  case(1):
    J1   = innerControl1Estimate(md,time,prices);    
    J    = innerControl1Exact(md,time,prices);
    J1V  = innerControl1Variance(md,time,prices);
    I1J1 = innerControl1Correlation(md,time,prices);
    break;
  case(2):
    J1 = innerControl2Estimate(md,time,prices); 
    J = innerControl2Exact(md,time,prices); 
    J1V  = innerControl2Variance(md,time,prices); 
    I1J1 = innerControl2Correlation(md,time,prices); 
    break;
  case(3):
    J1 = innerControl3Estimate(md,time,prices); 
    J = innerControl3Exact(md,time,prices); 
    J1V  = innerControl3Variance(md,time,prices); 
    I1J1 = innerControl3Correlation(md,time,prices); 
    break;
  }/* end switch */

  I1   = innerMeshEstimate(md,time,prices);
  beta = (I1J1 - I1 * J1)/(J1V - J1 * J1);
  return(I1 - beta*(J1 - J));
}

void innerDataFill(meshData *md, int whichControl){
  int time, path, toPath, fromPath;

  for(time=md->option.pData.times-2; time>=0; time--){
    for(toPath=0; toPath<md->meshPaths; toPath++){
      md->distribution[time][toPath] = 0;
      for(fromPath=0; fromPath<md->meshPaths;
          fromPath+=md->option.pData.clusterSize){
        md->distribution[time][toPath] += md->option.pData.clusterSize*
          pathTransitionProb(&(md->option.pData), time, fromPath, toPath);
      }
    }
  }

  for(time=md->option.pData.times-2;time>0;time--){
    for(path=md->option.pData.meshPaths-1;path>=0;path--){
      md->newValueIfHeld[time][path] = 
        innerNewEstimate(md,time,md->option.pData.price[time][path],
                         whichControl);
      if(md->newValueIfHeld[time][path] > md->option.payoff[time][path]){
        md->newValue[time][path] = md->newValueIfHeld[time][path];
      }else{
        md->newValue[time][path] = md->option.payoff[time][path];
      }
    }
  }

  md->newValueIfHeld[0][0] = 
    innerNewEstimate(md,0,md->option.pData.price[0][0],whichControl);
  if(md->newValueIfHeld[0][0] > md->option.payoff[0][0]){
    md->newValue[0][0] = md->newValueIfHeld[0][0];
  }else{
    md->newValue[0][0] = md->option.payoff[0][0];
  }
}


double innerControlledPathEstimate(meshData *md, int whichControl){
  int time, underlier, toPath, count;
  double **path, **antiPath, antiGrowth, *payoff, *antiPayoff, *valueIfHeld,
    *antiValueIfHeld, value, antiValue, weight, antiWeight, delta; 
  float *randomNum;

  allocate2dDouble(&path, md->option.pData.times, md->option.pData.underliers);
  allocate2dDouble(&antiPath,  md->option.pData.times,
                   md->option.pData.underliers);
  allocate1dDouble(&payoff, md->option.pData.times);
  allocate1dDouble(&antiPayoff, md->option.pData.times);
  allocate1dDouble(&valueIfHeld, md->option.pData.times);
  allocate1dDouble(&antiValueIfHeld, md->option.pData.times);


  value = 0;
  antiValue = 0;

  for(count=1; count<=md->option.pData.paths; count++){

    pathPair(&(md->option.pData),path,antiPath);

    for(time=0; time < md->option.pData.times; time++){
      payoff[time] = exp(-time*md->option.riskFreeRate)*
        optionPayoff(&(md->option), path[time]);
      antiPayoff[time] = exp(-time*md->option.riskFreeRate)*
        optionPayoff(&(md->option), antiPath[time]);
    }
      
    for(time=md->option.pData.times-2; time>=0; time--){
      valueIfHeld[time] = innerNewEstimate(md,time,path[time],whichControl);
      antiValueIfHeld[time] = innerNewEstimate(md,time,antiPath[time],
                                               whichControl);
    }

    for(time=0;time< md->option.pData.times;time++){
      if(payoff[time] > valueIfHeld[time]){
        value += payoff[time];
        break;
      }
    }

    for(time=0;time< md->option.pData.times;time++){
      if(antiPayoff[time] > antiValueIfHeld[time]){
        antiValue += antiPayoff[time];
        break;
      }
    }
  }

  value /= md->option.pData.paths;
  antiValue /= md->option.pData.paths;

  deallocate2dDouble(&path,  md->option.pData.times,
                     md->option.pData.underliers); 
  deallocate2dDouble(&antiPath,  md->option.pData.times,
                     md->option.pData.underliers); 
  deallocate1dDouble(&payoff, md->option.pData.times); 
  deallocate1dDouble(&antiPayoff, md->option.pData.times); 
  deallocate1dDouble(&valueIfHeld, md->option.pData.times); 
  deallocate1dDouble(&antiValueIfHeld, md->option.pData.times); 


  return(0.5*(value+antiValue));
}
